"""
Classes to represent Wordle words we want to guess and our guesses.
"""

from abc import ABC
from enum import Enum, auto
from typing import TYPE_CHECKING, List, Tuple

if TYPE_CHECKING:
    from wordle_benchmark.dictionary.wordle_dictionary import Dictionary


class LetterState(Enum):
    """ Enum of possible Wordle letter states """

    GREEN = auto()
    YELLOW = auto()
    BLACK = auto()


MatchState = Tuple[str, LetterState]


class Word(ABC):
    """ Word ABC """

    def __init__(self, word: str) -> None:
        """
        Args:
            word: Guess word or target word.
        """

        self._word = word.lower()

    def __contains__(self, value):
        return value in self._word

    def __iter__(self):
        return self._word.__iter__()

    def __len__(self) -> int:
        return len(self._word)

    def __str__(self) -> str:
        return self._word

    def in_dictionary(self, dictionary: "Dictionary") -> bool:
        """
        Check if a guess is in a dictionary.

        Args:
            dictionary: Dictionary to check membership in.
        """

        return self._word in dictionary

    def is_valid(self, word_len: int) -> bool:
        """
        Check if a guess is valid given board length constraint.

        Args:
            word_len: Length of word.
        """

        if not str(self).isalpha():
            return False

        if len(self) != word_len:
            return False

        return True


class TargetWord(Word):  # pylint: disable=too-few-public-methods
    """ Word we are trying to guess """

    ...


class GuessWord(Word):
    """ Word we have guessed """

    def compare_to(self, target_word: TargetWord) -> List[MatchState]:
        """
        Check matching letters between the guess and the target word.

        Args:
            target_word: Actual word being guessed.
        """

        match_list: List[MatchState] = []

        for self_letter, target_letter in zip(self, target_word):
            if self_letter == target_letter:
                match_list.append((self_letter, LetterState.GREEN))
            elif self_letter in target_word:
                match_list.append((self_letter, LetterState.YELLOW))
            else:
                match_list.append((self_letter, LetterState.BLACK))

        return match_list



"""
Classes to represent game logic.
"""

import logging
from enum import Enum, auto
from typing import TYPE_CHECKING, Generator, List, Optional, Tuple

from wordle_benchmark.dictionary.wordle_dictionary import RemoteDictionary
from wordle_benchmark.game.wordle_words import GuessWord, LetterState, TargetWord

if TYPE_CHECKING:

    from wordle_benchmark.dictionary.wordle_dictionary import Dictionary
    from wordle_benchmark.game.wordle_words import MatchState

log = logging.getLogger(__name__)


DEFAULT_DICTIONARY = RemoteDictionary(
    "https://raw.githubusercontent.com/dwyl/english-words/master/words_dictionary.json",  # pylint: disable=line-too-long
    word_len=5,
    seed=42,
)


class GameState(Enum):
    """ Enum of possible Wordle game states """

    UNSTARTED = auto()
    STARTED = auto()
    FINISHED = auto()


class IllegalGuessError(Exception):
    """ Guess is not the right length or is not a word """


class Game:  # pylint: disable=too-many-instance-attributes
    """ Logic for game """

    def __init__(
        self,
        target_word: str,
        dictionary: "Dictionary" = DEFAULT_DICTIONARY,
        max_guesses: int = 6,
    ):
        """
        Args:
            word: target word.
            dictionary: Wordle dictionary.
            max_guesses: max number guesses.
        """

        self._target_word = TargetWord(target_word)
        self._dictionary = dictionary
        self._max_guesses = max_guesses
        self._word_len = len(target_word)
        self._guesses: List[List["MatchState"]] = []

        self._greens: List[Tuple[str, int]] = []
        self._yellows: List[Tuple[str, int]] = []
        self._blacks: List[str] = []

        self._game_state = GameState.UNSTARTED

    @property
    def last_match(self) -> Optional[List["MatchState"]]:
        """ Most recent match. """

        if len(self._guesses) == 0:
            return None

        return self._guesses[-1]

    @property
    def n_guesses(self) -> int:
        """ Number of guesses so far. """

        return len(self._guesses)

    @property
    def possible_words(self) -> List[str]:
        """ Words that are still legal based on the dictionary. """

        possible_words = []

        for word in self._dictionary.word_list:

            # remove a word if it has a black character
            if any(character in word for character in self._blacks):
                continue

            if not all(word[ind] == character for character, ind in self._greens):
                continue

            if not all(character in word for character, _ in self._yellows):
                continue

            if any(word[ind] == character for character, ind in self._yellows):
                continue

            possible_words.append(word)

        return possible_words

    @property
    def success(self) -> bool:
        """ Whether the most recent guess was correct. """

        last_match = self.last_match

        if last_match is None:
            return False

        return all(letter_state == LetterState.GREEN for _, letter_state in last_match)

    def _transition_to_started(self):
        """ State machine transition to started """

        assert self._game_state == GameState.UNSTARTED

        log.info("Starting game...")
        self._game_state = GameState.STARTED

    def _transition_to_finished(self):
        """ State machine transition to finished. """

        assert self._game_state == GameState.STARTED

        log.info("Ending game")
        self._game_state = GameState.FINISHED

    def _handle_match(self, match_state: "MatchState", ind: int) -> None:
        """
        Update knowledge of black, yellow, green letters after guess.
        A match state represents a letter and it's black, yellow, green state.

        Args:
            match_state: Tuple[str, LetterState].
            ind: Letter position in guess word.
        """

        character, letter_state = match_state

        assert character.isalpha(), "Character must be standard alphabet"
        assert len(character) == 1, "A character should be length one"

        if letter_state == LetterState.BLACK:
            self._blacks.append(character)

        elif letter_state == LetterState.YELLOW:
            self._yellows.append((character, ind))

        elif letter_state == LetterState.GREEN:
            self._greens.append((character, ind))

        else:
            raise ValueError("Unknown letter state")

    def _register_guess(self, guess_word: "GuessWord") -> bool:
        """
        Register guess and update game state if necessary.

        Args:
            guess_word: Word guessed by user.
        Raises:
            IllegalGuessError in the event of an illegal word.
        """

        if not guess_word.is_valid(self._word_len):
            raise IllegalGuessError(
                f"{guess_word} is not valid word of len {self._word_len}"
            )

        if not guess_word.in_dictionary(self._dictionary):
            raise IllegalGuessError(f"{guess_word} is not in the dictionary")

        # black, yellow, green match outcome given guess
        comparison: List["MatchState"] = guess_word.compare_to(self._target_word)
        self._guesses.append(comparison)

        for ind, match in enumerate(comparison):
            self._handle_match(match, ind)

        # if all characters matched, that's the end of the game
        return self.success

    def start_game(self) -> Generator[None, str, None]:
        """
        Initiate game, await user input.
        NOTE: should be refactored to async await.

        Sample use of a manual game:
            game = Game(word, max_guesses=max_guesses)
            session = game.start_game()
            next(session)

            while True:
                session.send(input())

        Returns:
            A generator. Use ".send()" to pass guesses.
        """

        self._transition_to_started()

        log.debug("The target word is %s", self._target_word)

        while not self._game_state == GameState.FINISHED:

            log.info("Waiting for guess...")
            guess_word_str = yield
            log.info("Received guess %s", guess_word_str)
            guess_word = GuessWord(guess_word_str)

            try:
                end_game: bool = self._register_guess(guess_word)
            except IllegalGuessError as exception:
                log.warning(exception)
            else:
                if end_game:
                    self._transition_to_finished()
                    log.info("Correct! The word was %s", self._target_word)
                    break

                log.debug("These remain possible: %s", self.possible_words)

                if len(self._guesses) > self._max_guesses:
                    self._transition_to_finished()
                    log.info("Uh oh! You lose!")
                    break



"""
Classes to measure Wordle agent performance.
"""

from dataclasses import dataclass
import logging
import time
from typing import TYPE_CHECKING, List

import numpy as np

from wordle_benchmark.game import Game

if TYPE_CHECKING:
    from wordle_benchmark.agent import Agent

log = logging.getLogger(__name__)


@dataclass
class BenchmarkResults:
    """ Quantitative summarization of game outcomes """

    average_n_turns: float
    average_turn_time: float
    percent_successes: float
    std_turn_time: float


class Benchmark:  # pylint: disable=too-few-public-methods
    """ Performance measuring logic """

    def __init__(self, agent: "Agent", target_words: List[str], **game_kwargs):
        """
        Args:
            agent: Wordle playing agent.
            target_words: List of target words.
            game_kwargs: Optional game config keyword arguments.
        """

        self._agent = agent
        self._target_words = target_words
        self._game_kwargs = game_kwargs

    def run_games(self) -> BenchmarkResults:
        """ Run games against agent and record results """

        successes: List[bool] = []
        n_turns: List[int] = []
        turn_times: List[float] = []

        for target_word in self._target_words:

            log.info('Playing game with target word "%s"', target_word)

            game = Game(target_word, **self._game_kwargs)

            session = game.start_game()
            next(session)

            while True:

                try:

                    start = time.time()
                    session.send(self._agent.play(game))
                    end = time.time()

                    turn_times.append(end - start)

                except StopIteration:
                    break

            n_turns.append(game.n_guesses)
            successes.append(game.success)

        average_n_turns = np.mean(n_turns)
        average_turn_time = np.mean(turn_times)
        percent_successes = np.sum(successes) / len(successes)
        std_turn_time = np.std(turn_times)

        return BenchmarkResults(
            average_n_turns=average_n_turns,
            average_turn_time=average_turn_time,
            percent_successes=percent_successes,
            std_turn_time=std_turn_time,
        )



"""
Classes to represent Wordle playing agents.
"""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from wordle_benchmark.game import Game


class Agent(ABC):  # pylint: disable=too-few-public-methods
    """ Wordle agent """

    @abstractmethod
    def play(self, game: "Game") -> str:
        """
        Args:
            game: Game with game state information.
        Returns:
            String of guess word.
        """

        ...



"""
Classes to represent English words.
"""

import json
import pathlib
import random
from abc import ABC
from typing import List, Optional

import requests


class Dictionary(ABC):
    """ Dictionary ABC """

    def __init__(
        self, word_list: List[str], word_len: int, seed: Optional[int] = None
    ) -> None:
        """
        Args:
            word_list: list of all valid words.
            word_len: filter to apply to word length.
            seed: optional randomization seed to apply.
                NOTE! The default value is None and will result in no shuffle!
        """

        self._word_len = word_len
        self._word_list = word_list

        self._filtered_word_list = Dictionary._filter_by_len(word_list, word_len)
        self._randomized_order = Dictionary._randomize(self._filtered_word_list, seed)

    def __contains__(self, value):
        return self.word_list.__contains__(value)

    def __iter__(self):
        return self.word_list.__iter__()

    @property
    def word_len(self) -> int:  # pylint: disable=missing-function-docstring
        return self._word_len

    @property
    def word_list(self) -> List[str]:  # pylint: disable=missing-function-docstring
        return self._randomized_order.copy()

    @staticmethod
    def _filter_by_len(word_list: List[str], word_len: int) -> List[str]:
        """
        Filter to words of a specific length.
        NOTE: This really should be parallelized.

        Args:
            word_list: list of all valid words.
            word_len: filter to apply to word length.
        Returns:
            Words of length word_len.
        """

        return [word for word in word_list if len(word) == word_len]

    @staticmethod
    def _randomize(word_list: List[str], seed: Optional[int]) -> List[str]:
        """
        Randomize dictionary word list based on optional seed.
        NOTE: returns a copy, no inplace modification.

        Args:
            word_list: list of all valid words.
            seed: optional randomization seed to apply.
                NOTE! The default value is None and will result in no shuffle!
        Returns:
            Generate randomized dictionary order based on seed.
        """

        word_list_copy = word_list.copy()

        if seed is None:
            return word_list_copy

        random.Random(seed).shuffle(word_list_copy)

        return word_list_copy


class CustomDictionary(Dictionary):  # pylint: disable=too-few-public-methods
    """ CustomDictionary from lists """

    def __init__(self) -> None:  # pylint: disable=super-init-not-called
        raise NotImplementedError


class LocalDictionary(Dictionary):  # pylint: disable=too-few-public-methods
    """ LocalDictionary from local file """

    def __init__(self) -> None:  # pylint: disable=super-init-not-called
        raise NotImplementedError


class RemoteDictionary(Dictionary):  # pylint: disable=too-few-public-methods
    """ RemoteDictionary from remote source """

    def __init__(
        self, remote_source: str, word_len: int, seed: Optional[int] = None
    ) -> None:
        """
        Args:
            remote_source: remote URL for dictionary.
            word_len: filter to apply to word length.
            seed: optional randomization seed to apply.
                NOTE! The default value is None and will result in no shuffle!
        """

        word_list = RemoteDictionary._get_remote_source(remote_source)

        super().__init__(word_list=word_list, word_len=word_len, seed=seed)

    @staticmethod
    def _get_remote_source(remote_source: str) -> List[str]:
        """
        Download data from remote dictionary file.

        Args:
            remote_source: URL of remote.
        Returns:
            Word list.
        """

        return {".json": RemoteDictionary._list_from_json(remote_source)}[
            pathlib.Path(remote_source).suffix
        ]

    @staticmethod
    def _list_from_json(remote_source: str) -> List[str]:
        """
        Download JSON dictionary with key for words, arbitrary values.

        Args:
            remote_source: URL of remote JSON.
        Returns:
            Word list.
        """

        text = json.loads(requests.get(remote_source).text)

        return [word.lower() for word in sorted(text.keys()) if word.isalpha()]



"""
Call "pip install ." to initialize
Call "pip install -e ." for local development, see also requirements.txt
"""

import pathlib
import setuptools

README = (
    pathlib.Path(__file__).parent / "README.md"  # pylint: disable=unspecified-encoding
).read_text()

setuptools.setup(
    name="wordle-benchmark",
    version="1.0.2",
    description="Wordle package to benchmark agents. The package is intended \
        to support replicable heuristic evaluation of Wordle playing software.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Peter Bryan",
    author_email="peterbbryan@gmail.com",
    license="MIT",
    python_requires=">3.7",
    packages=setuptools.find_packages(),
    install_requires=["numpy", "requests"],
)
